/*
 * Simple inetd server.
 *
 * simple-inetd <port> <uid> <gid> <dir> <command> [<arg0>...]
 *
 * Example:
 * simple-inetd 10000 1001 1001 /home/user print-argv print-argv abc def
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>

/* #define HANDOVER_ENVVAL */
#define CONNECT_STDERR_TO_SOCKET
#define NO_CHANGE_ROOTDIR
#define NO_CHANGE_USER

typedef struct {
  int uid;
  int gid;
  char *dir;
  char *command;
  int argc;
  char **argv;
  char **envp;
} param_t;

static int create_socket(int port)
{
  int s, ret, on = 1;
  struct sockaddr_in address;

  s = socket(AF_INET, SOCK_STREAM, 0);
  if (s < 0) {
    perror("socket()");
    exit(-1);
  }

  ret = setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
  if (ret < 0) {
    perror("setsockopt()");
    exit(-1);
  }

  memset(&address, 0, sizeof(address));
  address.sin_family = AF_INET;
  address.sin_addr.s_addr = htonl(INADDR_ANY);
  address.sin_port = htons(port);

  ret = bind(s, (struct sockaddr *)&address, sizeof(address));
  if (ret < 0) {
    perror("bind()");
    exit(-1);
  }

  ret = listen(s, 5);
  if (ret < 0) {
    perror("listen()");
    exit(-1);
  }

  return s;
}

static int writestr(int s, char *buf)
{
  write(s, buf, strlen(buf));
  return 0;
}

static int command_exec(param_t *p)
{
  if (chdir(p->dir) < 0) {
    writestr(1, "fail chdir.\n");
    _exit(1);
  }

#ifndef NO_CHANGE_ROOTDIR
  if (chroot(".") < 0) {
    writestr(1, "fail chroot.\n");
    _exit(1);
  }

  if (chdir("/") < 0) {
    writestr(1, "fail chdir.\n");
    _exit(1);
  }
#endif

#ifndef NO_CHANGE_USER
  if (setgid(p->gid) < 0) {
    writestr(1, "fail setgid.\n");
    _exit(1);
  }

  if (setuid(p->uid) < 0) {
    writestr(1, "fail setuid.\n");
    _exit(1);
  }
#endif

  if (execve(p->command, p->argv, p->envp) < 0) {
    writestr(1, "fail exec.\n");
    _exit(1);
  }

  return 0;
}

static int server_proc(int s, param_t *p)
{
  int pid;

  pid = fork();
  if (pid < 0) {
    writestr(s, "fail fork.\n");
    exit(1);
  }
  if (pid) { /* parent */
    return 0;
  }

  /* child process */
  dup2(s, 0);
  dup2(s, 1);
#ifdef CONNECT_STDERR_TO_SOCKET
  dup2(s, 2);
#endif

  command_exec(p);
  _exit(0);
}

static void sigchld_handler(int val)
{
  int status;
  wait(&status);
}

static int server_main(int s, param_t *p)
{
  int ret;
  int sockt;
  struct sockaddr_in address;
  socklen_t len;

  signal(SIGCHLD, sigchld_handler);

  while (1) {
    memset(&address, 0, sizeof(address));
    len = sizeof(struct sockaddr_in);
    sockt = accept(s, (struct sockaddr *)&address, &len);
    if (sockt < 0) {
      continue;
    }
    ret = server_proc(sockt, p);
    close(sockt);
    if (ret < 0)
      break;
  }

  return 0;
}

static int help()
{
  fprintf(stderr, "simple-inetd <port> <uid> <gid> <dir> <command> [<arg0>...]\n");
  exit(1);
}

int main(int argc, char *argv[], char *envp[])
{
  int port, s, uid;
  param_t param;
#ifndef HANDOVER_ENVVAL
  char *null_args[] = { NULL };
#endif

  uid = getuid();
#if !defined(NO_CHANGE_ROOTDIR) || !defined(NO_CHANGE_USER)
  if (uid != 0) {
    fprintf(stderr, "You need execute this program as root.\n");
    exit(1);
  }
#endif

  if (argc < 6) {
    help();
  }
  argc--; argv++;

  port          = atoi(*argv); argc--; argv++;
  param.uid     = atoi(*argv); argc--; argv++;
  param.gid     = atoi(*argv); argc--; argv++;
  param.dir     = *argv;       argc--; argv++;
  param.command = *argv;       argc--; argv++;
  param.argc = argc;
  param.argv = argv;

#ifdef HANDOVER_ENVVAL
  param.envp = envp;
#else
  param.envp = null_args;
#endif

  s = create_socket(port);
  server_main(s, &param);

  return 0;
}
